import numpy as np
cimport numpy as np
from ridge.racer import (
    predictors,
    link_functions
)


# Gradient Steps of 1 epoch.

def mf(np.ndarray ratings, np.ndarray P, np.ndarray Q, float alpha):
    # [START Variable definitions]
    errors = []
    cdef int k = P.shape[0]
    cdef long int n_users = ratings.shape[0]
    cdef long int n_items = ratings.shape[1]
    cdef float residue
    # [END Variable definitions]

    # [START Calculation]
    for i in range(n_users):
        for j in np.nonzero(ratings[i, :])[0]:
            p = P[:, i]
            q = Q[:, j]
            residue = ratings[i, j] - np.dot(p, q)
            for f in range(k):
                P[f, i] += alpha * (2 * residue * Q[f][j])
                Q[f, j] += alpha * (2 * residue * P[f][i])
            errors.append(residue)
    # [END Calculation]
    
    return (P, Q, errors)

def fm_classification(float b, np.ndarray w, np.ndarray V, int k,
                      float l2, float eta, np.ndarray X, np.ndarray y):
    """Gradient Steps for FM Classification Task.

    TODO Too Slow !!!
    """
    # Variable Definition.
    cdef int m, f, i
    cdef float obs, coef, shared_term
    cdef np.ndarray row, sample_indices, nnz_ind
    cdef float b_new = b
    cdef np.ndarray w_new = w
    cdef np.ndarray V_new = V
    # Calculation of gradients.
    sample_indices = np.arange(start=0, stop=len(y))
    np.random.shuffle(sample_indices)
    for m in sample_indices:
        row = X[m]
        obs = y[m]
        # Parameter Updation
        z = predictors.fm(b, w, V, row)
        proba = link_functions.sigmoid(z)
        coef = proba - obs
        b_new -= eta * (coef + l2 * b)
        w_new -= eta * (coef * row + l2 * w)
        for f in np.arange(start=0, stop=k):
            nnz_ind = np.nonzero(row)[0]
            shared_term = np.sum([V[j,f] * row[j] for j in nnz_ind])
            for i in nnz_ind:
                first_term = row[i] * shared_term
                second_term = V[i,f] * np.square(row[i])
                V_new[i,f] -= eta * (coef * (first_term - second_term) + l2 * V_new[i,f])

    return (b_new, w_new, V_new)



# -*- coding: utf-8 -*-
import numpy as np
from scipy import stats

# stats.norm.pdf(x, loc=0, scale=1)


class SoftRank:
    """SoftRank: Optimizing Non-Smooth Rank Metrics (2008).

    モデルの出力は，エンティティのリスト(とそれを生成する確率)．
    その際の入力は，エンティティの素性のリスト(行列？)．
    """

    def __init__(self):
        pass

    def predict(self, X):
        """Prediction with a given features.

        Parameters
        ----------
        X : np.ndarray, whose shape is (n_entities, n_features)
        """



import numpy as np
cimport numpy as np
from ridge.racer import predictors
from ridge.racer import link_functions


# Calculation of loss, depending on the task of the model.

def fm_cross_entropy(float b, np.ndarray w, np.ndarray V,
                       np.ndarray X, np.ndarray y):
    """Given X and y, calculate an `approx` cross entropy loss.

    Parameters
    ----------
    X : np.ndarray, whose shape is (n_samples, n_features).
    y : np.ndarray of {0, +1}, whose shape is (n_samples, ).
    """
    # Variable Definitions
    cdef float loss = 0.0
    cdef float score, proba
    cdef np.ndarray y_nnz_ind = np.nonzero(y)[0]
    cdef np.ndarray y_zero_ind = np.where(y == 0)[0]

    # Loss Calculation
    for i in y_nnz_ind:
        score = predictors.fm(b, w, V, X[i])
        proba = link_functions.sigmoid(score)
        loss -= np.log(proba)
    for j in y_zero_ind:
        score = predictors.fm(b, w, V, X[i])
        proba = link_functions.sigmoid(score)
        loss -= np.log(1.0 - proba)
    
    return loss



import numpy as np
cimport numpy as np
from cpython cimport array
import array


# Prediction function of models.

def fm(float b, np.ndarray w, np.ndarray V, np.ndarray x):
    """Prediction of FM for Regression Task.

    Parameters
    ----------
    b : float, the bias term of FM model.
    w : np.ndarray, its shape is (n_features, ).
        Point-wise weight term of FM model.
    V : np.ndarray, its shape is (n_features, k).
        A vector of latent vectors of pair-wise interaction.
    x : np.ndarray, its shape is (n_features, ).
    """
    # Variable Definition
    cdef int k = V.shape[1]
    cdef np.ndarray nnz_ind = np.nonzero(x)[0]
    cdef float pointwise_score = np.dot(w[nnz_ind], x[nnz_ind])
    cdef float pairwise_score
    cdef array.array pairwise_scores = array.array('f', [])
    cdef float score
    # Within for loops
    cdef int i, f
    cdef float x_i
    cdef float V_if
    cdef float agg_sum_squared
    cdef float agg_squared_sum
    cdef array.array sum_squared, squared_sum

    # Calculation of Pair-wise interactions.
    for f in np.arange(start=0, stop=k):
        sum_squared = array.array('f', [])
        squared_sum = array.array('f', [])
        for i in nnz_ind:
            x_i = x[i]
            V_if = V[i,f]
            sum_squared.append(V_if * x_i)
            squared_sum.append(np.square(x_i) * np.square(V_if))
        agg_sum_squared = np.square(np.sum(sum_squared))
        agg_squared_sum = np.sum(squared_sum)
        pairwise_scores.append(agg_sum_squared - agg_squared_sum)
    pairwise_score = np.sum(pairwise_scores)

    # Score predicted.
    score = b + pointwise_score + 0.5 * pairwise_score

    return score



# -*- coding: utf-8 -*-
import numpy as np
from setuptools import setup, find_packages, Extension
from Cython.Distutils import build_ext


path2src = 'ridge/racer/src/'
ext_modules = [
    Extension('ridge.racer.gradient_steps', [path2src + 'gradient_steps.pyx']),
    Extension('ridge.racer.link_functions', [path2src + 'link_functions.pyx']),
    Extension('ridge.racer.loss_calculators', [path2src + 'loss_calculators.pyx']),
    Extension('ridge.racer.predictors', [path2src + 'predictors.pyx']),
]

setup(
    name='ridge',
    version='0.0.5',
    description='Python Machine Learning Library specialized in L2R and Recommendation.',
    author='Moriaki Saigusa',
    author_email='moriaki3193@gmail.com',
    url='https://github.com/moriaki3193/ridge',
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'numpy',
        'scipy',
        'tqdm',
    ],
    cmdclass={'build_ext': build_ext},
    ext_modules=ext_modules,
    include_dirs=[np.get_include()],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
import numpy as np
from tqdm import tqdm
from numpy.linalg import norm
from ridge.racer import gradient_steps


class MatFac:
    """Matrix Factorization
    """

    def __init__(self):
        self.P = None
        self.Q = None
        self.loss_series = []

    def _calc_loss(self, errors, l2):
        """エポックごとの損失を計算する

        params
        ------
        errors : list of float
                A difference between observed and predicted values.

        variables
        ---------
        squared_error_sum : float 実測値と予測値の二乗誤差の総和
        reg_term : L2正則化項
        """
        squared_error_sum = sum([residue ** 2 for residue in errors])
        reg_term = (l2 / 2.0) * (norm(self.P) + norm(self.Q))
        loss = squared_error_sum + reg_term
        self.loss_series.append(loss)
        return loss

    def fit(self, ratings, k=8, n_iter=1000,
            alpha=0.0005, l2=0.02, threshold=0.001, verbose=True):
        """レーティングを受け取り，ユーザとアイテムの`k'次元の潜在ベクトルを獲得する

        params
        ------
        ratings : ndarray or matrix, whose shape is (n_users, n_items)
                A matrix expressing the ratings of items by users.
        k       : int
                The number of dimensions of latent vectors.
        n_iter  : int
                The number of iterations when fitting.
        alpha   : float
        l2      : float
                A L2 regularization term.
        threshold : float
        verbose : bool
                Whether it displays progress bar when fitting.
        """
        ratings = np.asarray(ratings)

        # [START Initialize latent matrices]
        n_users, n_items = ratings.shape
        mu = 0.0
        sigma = 0.01
        self.P = np.random.normal(mu, sigma, (k, n_users))
        self.Q = np.random.normal(mu, sigma, (k, n_items))
        # [END Initialize latent matrices]

        # [START Fitting]
        pbar = tqdm(total=n_iter) if verbose else None
        for _epoch in range(n_iter):
            self.P, self.Q, errors = gradient_steps.mf(ratings, self.P, self.Q, alpha)
            loss = self._calc_loss(errors, l2)
            if verbose:
                pbar.update(1)
            if loss < threshold:
                break
        if verbose:
            pbar.close()
        # [END Fitting]

        return self

    def predict_one(self, user_id, item_id):
        """self.fitの結果を利用して，ユーザのアイテムへのレーティングの予測値を出力する
        """
        p = self.P[:, user_id]
        q = self.Q[:, item_id]
        return np.dot(p, q)


class NNMatFac:
    """Non-Negative Matrix Factorization

    Parameters
    ----------
    M : int, a hyper-param, the number of dims of vector S_{d,n}.
    W : np.ndarray, a latent matrix, whose shape is (D, M).
    H : np.ndarray, a latent matrix, whose shape is (M, N).

    References
    ----------
    + https://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf
    + https://www.slideshare.net/DaichiKitamura/efficient-initialization-for-nonnegative-matrix-factorization-based-on-nonnegative-independent-component-analysis
    """

    def __init__(self):
        self.M = None
        self.W = None
        self.H = None
        self.loss_series = None

    def _initialize_params(self, V, M):
        """Initialization of parameters.

        Parameters
        ----------
        D : int, the number of dims of features.
        M : int, the number of dims of latent vectors.
        N : int, the number of dims of samples.
        """
        D, N = V.shape
        self.D = D
        self.M = M
        self.N = N
        # self.W = np.random.rand(D, M)
        # self.H = np.random.rand(M, N)
        self.W = np.ones((D, M)) * np.sqrt(np.mean(V.mean() / (D * N)))
        self.H = np.ones((M, N)) * np.sqrt(np.mean(V.mean() / (D * N)))

    def fit(self, V, M, n_iter=10, verbose=True):
        """Fitting: Minimize KL divergence.

        Parameters
        ----------
        V : np.ndarray or sparse.csr_matrix, whose shape is (D, N).
        M : int, a hyper-param, the number of dims of vector S_{d,n}.
        """
        self._initialize_params(V, M)
        pbar = None
        if verbose:
            pbar = tqdm(total=n_iter)
            pbar.set_description('Fit. NMF')
        # [START Decomposition]
        for _ in range(n_iter):
            W_cur = self.W  # current W, shape is (D, M)
            H_cur = self.H  # curretn H, shape is (M, N)
            WH_cur = np.dot(W_cur, H_cur)  # shape is (D, N)
            # [START Update H]
            for m in range(self.M):
                for n in range(self.N):
                    numer = np.sum(W_cur[:,m] * (V[:,n] / WH_cur[:,n]))
                    denom = np.sum(W_cur[:,m])
                    self.H[m,n] = H_cur[m,n] * (numer / denom)
            # [END Update H]
            # [START Update W]
            for d in range(self.D):
                for m in range(self.M):
                    numer = np.sum(H_cur[m,:] * (V[d,:] / WH_cur[d,:]))
                    denom = np.sum(H_cur[m,:])
                    self.W[d,m] = W_cur[d,m] * (numer / denom)
            # [END Update W]
            if verbose:
                pbar.update(1)
        # [END Decomposition]
        pbar.close()

        return self
        


# -*- coding: utf-8 -*-
import numpy as np
from tqdm import tqdm
from scipy import sparse
from ridge.racer import (
    predictors,
    link_functions,
    loss_calculators,
)


class FMRegressor:
    """2-way Factorization Machine.

    Parameters
    ----------
    k : int, the hyper-param of FM.
    b : float, the bias term.
    w : ndarray, the element-wise weight vector.
    V : ndarray, whose shape is (n_features, k).
        The weight matrix of pairwise interaction terms.
    l2 : float, L2 regularization term.
    eta : float, a.k.a learning rate.
    """

    def __init__(self, type_X=np.ndarray):
        self.type_X = type_X
        self.k = None
        self.b = None
        self.w = None
        self.V = None
        self.l2 = None
        self.eta = None

    def _extract_feature_row(self, X, m):
        if self.type_X == sparse.csr_matrix:
            return np.squeeze(np.asarray(X[m].todense()))
        elif self.type_X == np.ndarray:
            return X[m]

    def _initialize_params(self, n_features, k, l2):
        self.k = k
        self.l2 = l2
        self.b = 0.0
        self.w = np.zeros(n_features)
        self.V = np.random.normal(scale=1e-2, size=(n_features, k))

    def _update_params(self, x, residue):
        """Update parameters using Stochastic Gradient Descent method.
        """
        nonzero_ind = np.nonzero(x)[0]
        self.b = self.b + self.eta * (residue - self.l2 * self.b)
        self.w = self.w + self.eta * (residue * x - self.l2 * self.w)
        V_cur = self.V
        self.V -= self.l2 * self.V  # L2 Regularize in advance
        for f in range(self.k):
            shared_term = np.sum([V_cur[j, f] * x[j] for j in nonzero_ind])
            for i in nonzero_ind:
                x_i = x[i]
                V_if = V_cur[i, f]
                partial = x_i * shared_term - V_if * np.power(x_i, 2)
                self.V[i, f] = V_if + self.eta * residue * partial

    def __score(self, x):
        """A prediction with a given feature vector.

        Utilizing Original Factorization Machine.

        Parameters
        ----------
        x : ndarray, whose shape is (n_features, )

        Return
        ------
        y_pred : float, a predicted target value.
        """
        return predictors.fm(self.b, self.w, self.V, x)

    def predict(self, X_test):
        """A prediction with given feature vectors.

        Parameters
        ----------
        X_test : ndarray, whose shape is (n_tests, n_features).

        Return
        ------
        y_pred : ndarray, whose shape is (n_tests, ).
        """
        n_samples, _ = X_test.shape
        y_pred = []

        # [START Predict one by one]
        for m in range(n_samples):
            x = self._extract_feature_row(X_test, m)
            this_y_pred = self.__score(x)
            y_pred.append(this_y_pred)
        # [END Predict one by one]

        return np.array(y_pred)

    def fit(self, X_train, y_train, k=8, l2=0.02, eta=1e-3, n_iter=1000):
        """Parameter fitting.

        Parameters
        ----------
        X_train : sparse.csr_matrix or ndarray,
                  whose shape is (n_samples, n_features).
        y_train : ndarray, whose shape is (n_samples, ).
        n_iter  : The maximum number of iteration.
        """
        n_samples, n_features = X_train.shape
        self._initialize_params(n_features, k, l2)
        # [START Fitting]
        self.eta = eta
        sample_indices = [i for i in range(n_samples)]
        for _epoch in tqdm(range(n_iter)):
            np.random.shuffle(sample_indices)
            for m in sample_indices:
                features = self._extract_feature_row(X_train, m)
                obs = y_train[m]
                pred = self.__score(features)
                self._update_params(features, obs - pred)
        # [END Fitting]

        return self


##############################################################################


class FMClassifier:
    """Factorization Machines for Classification.

    Features of this model.
    -----------------------
    + L2 Regularization.
    + Probability prediction.
    + Memoize a series of loss.
    + (Locally) Optimized logistic loss.
    + Probabilistic Approximation of learning rate (a.k.a Robbins-Monro).
    """

    def __init__(self):
        self.k = None
        self.b = None
        self.w = None
        self.V = None
        self.l2 = None
        self.eta = None
        self.loss_series = None

    def _initialize_loss_series(self, n_iter):
        """Initialize a series of loss when fitting.
        """
        self.loss_series = np.zeros(n_iter)

    def _update_loss_series(self, epoch, X, y):
        """Append Cross Entropy Loss to `loss_series'.
        """
        loss = loss_calculators.fm_cross_entropy(self.b, self.w, self.V, X, y)
        self.loss_series[epoch] = loss

    def _initialize_params(self, n_features, k, l2, eta):
        """Initialize parameters of this model.
        """
        self.k = k
        self.b = 0.0
        self.w = np.zeros(n_features)
        self.V = np.random.normal(scale=1e-2, size=(n_features, k))
        self.l2 = l2
        self.eta = eta

    def _update_params(self, x, y):
        """Update parameters of this model.

        Employing Stochastic Gradient Descent algorithm
        with Robbins-Monro method.

        Parameters
        ----------
        x : np.ndarray, whose shape is (n_features, )
        y : int {0, +1}

        Variables
        ---------
        V_cur : Current values of latent matrix V.
        """
        z = self._predict_proba(x)
        coef = z - y
        self.b -= self.eta * (coef + self.l2 * self.b)
        self.w -= self.eta * (coef * x + self.l2 * self.w)
        V_cur = self.V
        for f in np.arange(start=0, stop=self.k):  # TODO 列方向の取り出しは低速なので改良する
            nnz_ind = np.nonzero(x)[0]
            shared_term = np.sum([V_cur[j, f] * x[j] for j in nnz_ind])
            for i in nnz_ind:
                first_term = x[i] * shared_term
                second_term = V_cur[i, f] * np.square(x[i])
                nabra = coef * (first_term - second_term)
                self.V[i, f] -= self.eta * (nabra + self.l2 * V_cur[i, f])

    def __score(self, x):
        """Scoring with a given feature `x'.

        Parameters
        ----------
        x : np.ndarray, whose shape is (n, ).

        Return
        ------
        score : float, regression score by FMs.
        """
        return predictors.fm(self.b, self.w, self.V, x)

    def _predict_proba(self, x):
        """Predict probability with a given feature `x'.

        A.k.a Sigmoid function.

        Parameters
        ----------
        x : np.ndarray, whose shape is (n_features, ).
        """
        return link_functions.sigmoid(self.__score(x))

    def fit(self, X, y, k=8, l2=1e-2, eta=1e-2, n_iter=1000, verbose=True):
        """Fitting parameters of this model.

        Parameters
        ----------
        X       : np.ndarray, whose shape is (n_samples, n_features).
        y       : np.ndarray, whose shape is (n_samples, ).
                  Each element in y must be either {0, +1}.
        k       : int, a hyper-param of this model. Default to 8.
        l2      : float, a L2 Regularization term. Default to 1e-2.
        eta     : float, an initial value of learning rate.
                  This eta is attenuated as the number of iteration increases.
        verbose : bool, whether display a progress bar of iteration.
        """
        n_samples, n_features = X.shape
        self._initialize_params(n_features, k, l2, eta)
        self._initialize_loss_series(n_iter)
        sample_indices = np.arange(start=0, stop=n_samples)
        if verbose:
            pbar = tqdm(total=n_iter)
            pbar.set_description(f'Fitting')
        for epoch in np.arange(start=0, stop=n_iter):
            np.random.shuffle(sample_indices)
            for m in sample_indices:
                row = X[m]
                obs = y[m]
                self._update_params(row, obs)
            self.eta = eta / (epoch + 1)
            self._update_loss_series(epoch, X, y)
            if verbose:
                pbar.update(1)
        pbar.close()
        return self

    def predict(self, X, target='0-1'):
        """Prediction of this model.

        Parameters
        ----------
        X : np.ndarray, whose shape is (n_samples, n_features).

        Return
        ------
        y_pred : np.ndarray, whose shape is (n_samples, ).
                 if `target' is 'probability', return estimated probabilities.
                 otherwise, return a series of {0, +1} values.
        """
        if target == 'probability':
            return np.array([self._predict_proba(x) for x in X])
        else:
            probas = np.array([self._predict_proba(x) for x in X])
            return np.where(probas >= 0.5, 1, 0)


##############################################################################
